#include "selectablelinesegmentgraphic.h"

#include <QBrush>
#include <QLinearGradient>
#include <QDebug>
#include <QRect>
#include <QPolygon>
#include <QGraphicsPathItem>

#include "math.h"

/**
  * \class SelectableLineSegmentGraphic
  *
  *
  The SelectableLineSegmentGraphic will start out as a short horizontal line.
  It will have a starting position given by the parent.
  It will have a default length.

   If the user grabs a corner and drags, the absolute scene position of
   that new mouse position and the anchored corner will be user to calculate a new length,
   and a rotation angle will be calculated from base-line (horizontal).

   The line will then be re-drawn with the new length and new angle.
  *
  */

SelectableLineSegmentGraphic::SelectableLineSegmentGraphic(  QPointF position): QGraphicsPolygonItem(),
        _outterborderColor(Qt::black),
        _pen(),
        _location(0,0),
        _dragStart(0,0),
        _gridSpace(10),
        _lineLength(200),
        _XcornerGrabBuffer(20),
        _YcornerGrabBuffer(20),
        _graphicsItemBoundingBoxWidth( _lineLength +(2 * _XcornerGrabBuffer)),
        _lineEnd_0( _XcornerGrabBuffer, _YcornerGrabBuffer),
        _lineEnd_1 ( _XcornerGrabBuffer + _lineLength  , _YcornerGrabBuffer),
        _cornerGrabbed(false),
        _selectRegion()
{

    this->setPos(position);

    _pen.setWidth(2);
    _pen.setColor(_outterborderColor);

    this->setAcceptHoverEvents(true);


    QPointF p1  (_lineEnd_0.x() - _XcornerGrabBuffer, _lineEnd_0.y() - _YcornerGrabBuffer);
    QPointF p2  (_lineEnd_0.x() + _XcornerGrabBuffer, _lineEnd_0.y() - _YcornerGrabBuffer);
    QPointF p3  (_lineEnd_0.x() + _XcornerGrabBuffer, _lineEnd_0.y() + _YcornerGrabBuffer);
    QPointF p4  (_lineEnd_0.x() - _XcornerGrabBuffer, _lineEnd_0.y() + _YcornerGrabBuffer);


    QPointF p5  (_lineEnd_1.x() - _XcornerGrabBuffer, _lineEnd_1.y() - _YcornerGrabBuffer);
    QPointF p6  (_lineEnd_1.x() + _XcornerGrabBuffer, _lineEnd_1.y() - _YcornerGrabBuffer);
    QPointF p7  (_lineEnd_1.x() + _XcornerGrabBuffer, _lineEnd_1.y() + _YcornerGrabBuffer);
    QPointF p8  (_lineEnd_1.x() - _XcornerGrabBuffer, _lineEnd_1.y() + _YcornerGrabBuffer);

    _selectRegion << p1 << p2  << p5 << p6 << p7 << p8 << p3 << p4 << p1;

    this->setPolygon(_selectRegion);

}




/**
  * This scene event filter has been registered with all four corner grabber items.
  * When called, a pointer to the sending item is provided along with a generic
  * event.  A dynamic_cast is used to determine if the event type is one of the events
  * we are interrested in.
  */
bool SelectableLineSegmentGraphic::sceneEventFilter ( QGraphicsItem * watched, QEvent * event )
{
    //qDebug() << " QEvent == " + QString::number(event->type());

    CornerGrabber * corner = dynamic_cast<CornerGrabber *>(watched);
    if ( corner == NULL) return false; // not expected to get here

    QGraphicsSceneMouseEvent * mevent = dynamic_cast<QGraphicsSceneMouseEvent*>(event);
    if ( mevent == NULL)
    {
        // this is not one of the mouse events we are interrested in
        return false;
    }


    switch (event->type() )
    {
        // if the mouse went down, record the x,y coords of the press, record it inside the corner object
    case QEvent::GraphicsSceneMousePress:
        {
            corner->setMouseState(CornerGrabber::kMouseDown);

            QPointF scenePosition =  corner->mapToScene(mevent->pos());
            corner->mouseDownY = scenePosition.y();
            corner->mouseDownX= scenePosition.x();

            _cornerGrabbed = true;

        }
        break;

    case QEvent::GraphicsSceneMouseRelease:
        {
            corner->setMouseState(CornerGrabber::kMouseReleased);
            _cornerGrabbed = false;
        }
        break;

    case QEvent::GraphicsSceneMouseMove:
        {
            corner->setMouseState(CornerGrabber::kMouseMoving );
        }
        break;

    default:
        // we dont care about the rest of the events
        return false;
        break;
    }


    if ( corner->getMouseState() == CornerGrabber::kMouseMoving )
    {
        createCustomPath(mevent->pos(), corner);

        this->update();
    }

    return true;// true => do not send event to watched - we are finished with this event
}



void SelectableLineSegmentGraphic::createCustomPath(QPointF mouseLocation, CornerGrabber* corner)
{

    qreal lineEndX = 0;
    qreal lineEndY = 0;
    qreal lineStartX = 0;
    qreal lineStartY = 0;

    qreal mouseX =0, mouseY = 0;

    QPointF scenePosition = corner->mapToScene(mouseLocation);
    mouseY = scenePosition.y();
    mouseX = scenePosition.x();



    // which corner needs to get moved?
    if ( corner == _corners[0])
    {

        lineStartY =  mouseY;
        lineStartX =  mouseX;
        lineEndY = this->mapToScene( _corners[1]->getCenterPoint()).y();
        lineEndX = this->mapToScene( _corners[1]->getCenterPoint()).x();

    }
    else
    {
        lineStartY = this->mapToScene( _corners[0]->getCenterPoint()).y();
        lineStartX = this->mapToScene( _corners[0]->getCenterPoint()).x();

        lineEndY =  mouseY;
        lineEndX =  mouseX;

    }


    _lineEnd_0.setX(  mapFromScene(lineStartX, lineStartY).x()  );
    _lineEnd_0.setY(  mapFromScene(lineStartX, lineStartY).y() );


    _lineEnd_1.setX(  mapFromScene(lineEndX,lineEndY ).x()  );
    _lineEnd_1.setY(  mapFromScene(lineEndX,lineEndY ).y());



    enclosePath(lineStartX, lineStartY,  lineEndX, lineEndY  );

    setCornerPositions();

}

void SelectableLineSegmentGraphic::enclosePath(qreal lineStartX,qreal lineStartY, qreal lineEndX, qreal lineEndY  )
{


    /* to draw the select region around the line,
        define a large box(rect p1,p2,p3,p4) around each end of the line.
        then define a box(rect p5,p6,p7,p8) that connects the two end boxes.
        now no matter the angle of the line, we can be sure that its
        surrounded by a symetrical region that will pick up the mouse.


        .    .             .    .
           X-----------------X
        .    .             .    .

        the squares have to connect at the two pairs of square corners that are closest.

       */

    QList<QPointF> pointsStart;
    QList<QPointF> pointsEnd;

    pointsStart.append( QPointF(lineStartX  - _XcornerGrabBuffer, lineStartY - _YcornerGrabBuffer));
    pointsStart.append(  QPointF(lineStartX + _XcornerGrabBuffer, lineStartY - _YcornerGrabBuffer));
    pointsStart.append(  QPointF(lineStartX + _XcornerGrabBuffer, lineStartY + _YcornerGrabBuffer));
    pointsStart.append(  QPointF(lineStartX - _XcornerGrabBuffer, lineStartY + _YcornerGrabBuffer));

    pointsEnd.append(  QPointF(lineEndX - _XcornerGrabBuffer, lineEndY - _YcornerGrabBuffer));
    pointsEnd.append(  QPointF(lineEndX + _XcornerGrabBuffer, lineEndY - _YcornerGrabBuffer));
    pointsEnd.append(  QPointF(lineEndX + _XcornerGrabBuffer, lineEndY + _YcornerGrabBuffer));
    pointsEnd.append(  QPointF(lineEndX - _XcornerGrabBuffer, lineEndY + _YcornerGrabBuffer));

    qreal minDistance = 0 ;
    qreal secondMinDistance = 0;
    QPointF p1 ;
    QPointF p2 ;
    QPointF p3 ;
    QPointF p4 ;
    int i1=0;
    int i2=0;

    for (int i =0 ; i < 4; i++)
    {
        for (int j = 0; j < 4 ; j++)
        {
            qreal d1 = pow(pointsStart[i].x() - pointsEnd[j].x(), 2.0)
                       + pow( pointsStart[i].y() - pointsEnd[j].y(), 2.0);
            if ( d1 > minDistance)
            {
                minDistance = d1;
                p1 = pointsStart[i];
                p2 = pointsEnd[j];
                i1 = i;
                i2 = j;
            }
        }
    }



    for (int i =0 ; i < 4; i++)
    {
        if ( i == i1) continue;

        for (int j = 0; j < 4 ; j++)
        {
            if ( j == i2 ) continue;

            qreal d1 = pow(pointsStart[i].x() - pointsEnd[j].x(), 2.0)
                       + pow( pointsStart[i].y() - pointsEnd[j].y(), 2.0);
            if ( d1 > secondMinDistance)
            {
                secondMinDistance = d1;
                p3 = pointsStart[i];
                p4 = pointsEnd[j];
            }
        }
    }

    _selectRegion.clear();

    // first enclose each end in its box


    _selectRegion << mapFromScene(p1) << mapFromScene(p3)  << mapFromScene(p4) << mapFromScene(p2) << mapFromScene(p1) ;

    this->setPolygon(_selectRegion);

    this->update();

}



// for supporting moving the box across the scene
void SelectableLineSegmentGraphic::mouseReleaseEvent ( QGraphicsSceneMouseEvent * event )
{
    event->setAccepted(true);
    _location.setX( ( static_cast<int>(_location.x()) / _gridSpace) * _gridSpace );
    _location.setY( ( static_cast<int>(_location.y()) / _gridSpace) * _gridSpace );
    this->setPos(_location);
}


// for supporting moving the box across the scene
void SelectableLineSegmentGraphic::mousePressEvent ( QGraphicsSceneMouseEvent * event )
{
    event->setAccepted(true);
    _dragStart = event->pos();
}


// for supporting moving the box across the scene
void SelectableLineSegmentGraphic::mouseMoveEvent ( QGraphicsSceneMouseEvent * event )
{    
    QPointF newPos = event->pos() ;
    _location += (newPos - _dragStart);
    this->setPos(_location);
}

// remove the corner grabbers

void SelectableLineSegmentGraphic::hoverLeaveEvent ( QGraphicsSceneHoverEvent * )
{
    _outterborderColor = Qt::black;

    _corners[0]->setParentItem(NULL);
    _corners[1]->setParentItem(NULL);

    delete _corners[0];
    delete _corners[1];
}


// create the corner grabbers

void SelectableLineSegmentGraphic::hoverEnterEvent ( QGraphicsSceneHoverEvent * )
{
    _outterborderColor = Qt::red;

    _corners[0] = new CornerGrabber(this,0);
    _corners[1] = new CornerGrabber(this,1);

    _corners[0]->installSceneEventFilter(this);
    _corners[1]->installSceneEventFilter(this);


    setCornerPositions();

}

void SelectableLineSegmentGraphic::setCornerPositions()
{
    int cornerWidth = (_corners[0]->boundingRect().width())/2;
    int cornerHeight = ( _corners[0]->boundingRect().height())/2;

    _corners[0]->setPos(_lineEnd_0.x() - cornerWidth, _lineEnd_0.y() - cornerHeight );
    _corners[1]->setPos(_lineEnd_1.x()- cornerWidth, _lineEnd_1.y() - cornerHeight );

     qDebug() << "corner[0].pos : " + QString::number(_corners[0]->pos().x()) + ", " + QString::number(_corners[0]->pos().y())
     << "; corner[1].pos : " + QString::number(_corners[1]->pos().x()) + ", " + QString::number(_corners[1]->pos().y());
}

#if 0
QRectF Transistion::boundingRect() const
{
    return QRectF(0,0,_graphicsItemBoundingBoxWidth,_height);
}
#endif



void SelectableLineSegmentGraphic::paint (QPainter *painter, const QStyleOptionGraphicsItem *i, QWidget *w)
{
    // diag line
//    _pen.setStyle(Qt::NoPen);
//    _pen.setColor(Qt::transparent);
//    painter->setPen(_pen);

//    QGraphicsPolygonItem::paint(painter, i, w );
    // end diag line

    // draw the line

    _pen.setStyle(Qt::SolidLine);
    painter->setPen(_pen);

    _pen.setColor(Qt::red);
     painter->drawLine(_lineEnd_0, _lineEnd_1);



}



void SelectableLineSegmentGraphic::mouseMoveEvent(QGraphicsSceneDragDropEvent *event)
{
    event->setAccepted(false);
}

void SelectableLineSegmentGraphic::mousePressEvent(QGraphicsSceneDragDropEvent *event)
{
    event->setAccepted(false);
}
